*******************************************************************************
*   PURPOSE:    WMsgBox will place the given message in a window at the given
*               coordinates (or centered if -1 was passed).  The message may be
*               broken into multi-lines by embedding a CHR(13) in the message
*               string.  Mimics function MsgBox() found in FOXTOOLS.FLL.
*               However, the following features are included:
*
*                      User-definable button options
*                      Option to return numeric value
*                      Option to specify message's font, size, and style
*                      Option to specify message font's color
*                      Option to specify justification of message in window
*
*   SYNTAX:     wmsgbox (<cmsg>, <ctitle>, <ctype>, <cprompt>, <nrow>, <ncol>,
*                       [<cjust>], [<cfont>], [<nsize>], [<cstyle>], [<cncolor>],
*                       [1])
*
*   PARAMETERS: <cmsg>      - The messages to display...each line separated
*                             by a carriage return, CHR(13).
*               <ctitle>    - The title for the message window. Defaults to
*                             "System Message"
*               <ctype>     - Type of ICON to display with dialog or the file
*                             name of the ICON to display or the field name
*                             of the ICON to display.  Icon must be 36 x 36
*                             in size.  The alias may be included with the
*                             field name.
*               <cprompt>   - The options to choose from. Pass empty string
*                             to display message without waiting for user
*                             input.
*               <nrow>      - The row where top of window starts. Pass -1
*                             to have window centered.
*               <ncol>      - The column where left of window starts. Pass
*                             -1 to have window centered.
*               <cjust>     - Optional justification of message. Default is
*                             left justification.
*               <cfont>     - Optional font type.
*               <nsize>     - Optional Size of cfont.
*               <cstyle>    - Optional Style of cfont.
*               <cncolor>   - Optional Color string for given font.  Will
*                             accept an color string that is valid for a
*                             @...SAY command.  A color scheme or RGB(...)
*                             color style may be passed as well.  For color
*                             scheme, pass the scheme number; for RGB(...)
*                             enclosed the RGB(...) specification in quotes.
*               <1>         - Optional, if a 1 is passed in, then function
*                             returns a number indicating which button was
*                             chosen.
*
*   RETURNS:    Character or Numeric.
*
*******************************************************************************
FUNCTION Wmsgbox

PARAMETERS  cmsg, ctitle, ctype, cprompt, nrow, ncol, ;
            cjust, cfont, nsize, cstyle, cncolor, nnumeric

PRIVATE lvisible,   cdevice,    retval,     nselect,    cbitmaps,;
        cwinsave,   cprinter,   cconsole,   ciconfile,  i,;
        nbuttons,   ndefault,   nbutwidth,  nbutsize,   nbutheight, dbfields_,;
        cbfont,     nbsize,     cbstyle,    ngap,       nmaxstr,    aclrset
PRIVATE nwwidth,    nwheight,   nmaxlen,    nmaxline,   nsayrow,    nicongap,;
        nwfheight,  nchoice,    amsg,       aprompt,    nsaycol,    lpicbut,;
        nmwfheight, nbfheight,  nbfwidth,   nwbwidth,   nloc1,      nloc2,;
        amsgclr

*
* Release message window if it already exists...and clear the keyboard
* buffer
*
RELEASE WINDOW wmsgbox
CLEAR TYPEAHEAD

IF EMPTY(cmsg)
    RETURN ""
ENDIF

*
* Declare some psuedo-constants so WMSGBOX() will recognize the type of
* icons and type of dialog to initiate.
*
#DEFINE K_LEFTJUST          "L"
#DEFINE K_CENTERJUST        "C"
#DEFINE K_RIGHTJUST         "R"

#DEFINE K_NORMALSTYLE       "N"
#DEFINE K_BOLDSTYLE         "B"
#DEFINE K_ITALICSTYLE       "I"
#DEFINE K_ITALICBOLDSTYLE   "BI"

#DEFINE K_OK                "OK"
#DEFINE K_CONTINUE          "C"
#DEFINE K_OKCANCEL          "OKC"
#DEFINE K_YESNO             "YN"
#DEFINE K_YESNOCANCEL       "YNC"
#DEFINE K_RETRYCANCEL       "RC"
#DEFINE K_ABORTRETRYIGNORE  "ARI"

#DEFINE K_ICONHAND          "HAND"
#DEFINE K_ICONQUESTION      "?"
#DEFINE K_ICONEXCLAMATION   "!"
#DEFINE K_ICONASTERISK      "*"
#DEFINE K_ICONINFORMATION   "INFO"
#DEFINE K_ICONSTOP          K_ICONHAND

#DEFINE K_SAYCOLOR          SCHEME(1, 1)
#DEFINE K_ICONGAP           10
#DEFINE K_ICONWIDTH         36
#DEFINE K_ICONHEIGHT        K_ICONWIDTH
#DEFINE K_RETURN            CHR(13)
#DEFINE K_MAXREADLEVEL      5
#DEFINE K_CONTACTNAME       "Bob Lucas"    

*
* Foreground Character Code Colors to use with WMsgBox()
*
#DEFINE K_BLACK             "{N}"
#DEFINE K_BLUE              "{B+}"
#DEFINE K_CYAN              "{BG+}"
#DEFINE K_DRKBLUE           "{B}"
#DEFINE K_DRKCYAN           "{BG}"
#DEFINE K_DRKGREEN          "{G}"
#DEFINE K_DRKMAGENTA        "{RB}"
#DEFINE K_DRKRED            "{R}"
#DEFINE K_DRKYELLOW         "{GR}"
#DEFINE K_DRKGRAY           "{N+}"
#DEFINE K_GRAY              "{W}"
#DEFINE K_GREEN             "{G+}"
#DEFINE K_MAGENTA           "{RB+}"
#DEFINE K_RED               "{R+}"
#DEFINE K_WHITE             "{W+}"
#DEFINE K_YELLOW            "{GR+}"
#DEFINE K_MAXSCHEME         24

*
* Save some environment variables to restore later
*
cdevice     = SET("DEVICE")
cconsole    = SET("CONSOLE")
cprinter    = SET("PRINTER")
cwinsave    = WOUTPUT()
lvisible    = WVISIBLE(cwinsave)

DIMENSION aclrset[K_MAXSCHEME]

FOR i = 1 TO K_MAXSCHEME
    aclrset[i] = SCHEME(i)
ENDFOR


SET PRINTER OFF
SET CONSOLE ON
SET DEVICE TO SCREEN

*
* Initialize button's font info...
*
cbfont  = "MS Sans Serif"
nbsize  = 8
cbstyle = "B"

*
* Initialize needed variables
*
lpicbut     = .F.
nbutheight  = 1.75
nrow        = IIF(TYPE("nrow") <> "N", -1, nrow)
ncol        = IIF(TYPE("ncol") <> "N", -1, ncol)
retval      = IIF(TYPE("nnumeric") <> "N" OR EMPTY(nnumeric), "", 0)
ctype       = IIF(TYPE("ctype")<>"C", K_ICONEXCLAMATION, ;
                UPPER(ALLTRIM(ctype)))
ctitle      = IIF(TYPE("ctype")<>"C" OR EMPTY(ctitle), ;
                "System Message", ALLTRIM(ctitle))
*
* Determine type of padding to use in displaying message.  Default to
* flush left.
*
IF (TYPE("cjust") <> [C]) OR EMPTY(cjust) ;
    OR (NOT INLIST(UPPER(ALLTRIM(cjust)), K_LEFTJUST, K_CENTERJUST, K_RIGHTJUST))

    cjust = K_LEFTJUST
ELSE
    cjust = UPPER(ALLTRIM(cjust))
ENDIF

*
* Determine the corresponding Icon to display with message
*
cbitmaps    = ""
ciconfile   = ""

DO CASE
    CASE ctype == K_ICONEXCLAMATION
        ciconfile =  IIF(FILE("EXCLAIM.BMP"), "EXCLAIM.BMP", "")
    CASE ctype == K_ICONHAND OR ctype == K_ICONSTOP
        ciconfile =  IIF(FILE("STOP.BMP"), "STOP.BMP", "")
    CASE ctype == K_ICONINFORMATION OR ctype == K_ICONASTERISK
        ciconfile =  IIF(FILE("INFO.BMP"), "INFO.BMP", "")
    CASE ctype == K_ICONQUESTION
        ciconfile = IIF(FILE("QUESTION.BMP"), "QUESTION.BMP", "")
    OTHERWISE
        IF FILE(ctype)
            ciconfile   = ctype
        ELSE
            nloc1       = AT(".", ctype)
            nloc2       = AT("->", ctype)
            *
            * Save current work area and check if an alias was provided.
            *
            nselect = SELECT(0)
            DO CASE
                CASE (nloc1 > 0) OR (nloc2 > 0)
                    *
                    * If so, then parse string into alias and field parts.
                    *
                    cbitmaps    = LEFT(ctype, nloc2 - 1)
                    ciconfile   = SUBSTR(ctype, nloc2 + 2)
                    IF EMPTY(cbitmaps)
                        cbitmaps    = LEFT(ctype, nloc1 - 1)
                        ciconfile   = SUBSTR(ctype, nloc1 + 1)
                    ENDIF

                CASE NOT EMPTY(ALIAS())
                    cbitmaps    = ALIAS()
                    ciconfile   = ctype
            ENDCASE

            IF USED(cbitmaps)
                ctype = cbitmaps + "->" + ciconfile
                SELECT (cbitmaps)

                *
                * Verify that the field exists...
                *
                i = FCOUNT()
                DIMENSION dbfields_[i]
                = AFIELDS(dbfields_)

                ciconfile = IIF(ASCAN(dbfields_, ciconfile) > 0, ctype, "")
                RELEASE dbfields_

                SELECT (nselect)
                cbitmaps = "DBF"
            ELSE
                cbitmaps = ""
                ciconfile = ""
            ENDIF
        ENDIF
ENDCASE

*
* Initialize message window's font info and save old color setting...
*
cfont   = IIF(TYPE("cfont") <> "C" OR EMPTY(cfont), ;
            "MS Sans Serif", ALLTRIM(cfont))
nsize   = IIF(TYPE("nsize") <> "N" OR EMPTY(nsize), ;
            8, nsize)
cstyle  = IIF(TYPE("cstyle") <> "C" OR EMPTY(cstyle), ;
            "B", UPPER(ALLTRIM(cstyle)))

DO CASE
    CASE EMPTY(cncolor)
        cncolor = K_SAYCOLOR
    CASE TYPE("cncolor") == "N"
        cncolor = IIF(NOT BETWEEN(cncolor, 1, 12), K_SAYCOLOR, cncolor)
    CASE TYPE("cncolor") == "C"
        cncolor = ALLTRIM(cncolor)
    OTHERWISE
        cncolor = K_SAYCOLOR
ENDCASE

*
* Initialize window with the message font's properties so our work will
* depend on this window's font attributes and window color.
*
DEFINE WINDOW wmsgbox AT 0,0 SIZE 25,25 FONT cfont,nsize STYLE cstyle
ACTIVATE WINDOW wmsgbox NOSHOW

IF TYPE("cncolor") == "N"
    SET COLOR OF SCHEME 1 TO SCHEME cncolor
ELSE
    SET COLOR OF SCHEME 1 TO (cncolor)
ENDIF

*
* Height of the main window's font, the message window's font, and the
* button's font.
*
nmwfheight  = FONTMETRIC(1) + FONTMETRIC(5)
nwfheight   = FONTMETRIC(1, cfont, nsize, cstyle) ;
            + FONTMETRIC(5, cfont, nsize, cstyle)
nbfheight   = FONTMETRIC(1, cbfont, nbsize, cbstyle) ;
            + FONTMETRIC(5, cbfont, nbsize, cbstyle)
*
* Define icon width and height based on message window's font.
*
IF NOT EMPTY(ciconfile)
    niconheight = K_ICONHEIGHT / nwfheight
    niconwidth  = K_ICONWIDTH  / FONTMETRIC(6, cfont, nsize, cstyle)
    nicongap    = K_ICONGAP    / FONTMETRIC(6, cfont, nsize, cstyle)
ELSE
    niconheight = 0
    niconwidth  = 0
    nicongap    = 0
ENDIF

*
* Determine the longest string to display.  Each new line is indicated by
* a CHR(13), or K_RETURN, code.
*
nmaxline    = 1
nloc        = AT(K_RETURN, cmsg)

IF nloc > 0
    DO WHILE nloc > 0
        DIMENSION amsg[nmaxline]
        DIMENSION amsgclr[nmaxline]
        amsg[nmaxline] = get_msg(LEFT(cmsg, nloc-1), @amsgclr, nmaxline)

        nmaxline = nmaxline + 1
        cmsg = IIF(nloc+1 > LEN(cmsg), "", SUBSTR(cmsg, nloc+1))
        nloc = AT(K_RETURN, cmsg)
    ENDDO

    DIMENSION amsg[nmaxline]
    DIMENSION amsgclr[nmaxline]
    amsg[nmaxline] = get_msg(cmsg, @amsgclr, nmaxline)

    nmaxlen = -1
    FOR i = 1 TO nmaxline
        IF nmaxlen < LEN(amsg[i])
            nmaxlen = LEN(amsg[i])
            nmaxstr = i
        ENDIF
    ENDFOR
ELSE
    DIMENSION amsg[1]
    DIMENSION amsgclr[1]
    amsg[1]     = get_msg(cmsg, @amsgclr, 1)
    nmaxstr     = 1
    nmaxline    = 1
    nmaxlen     = LEN(cmsg)
ENDIF

*
* Parse the prompt statement if any
*
IF EMPTY(cprompt) OR TYPE("cprompt") <> "C"
    cprompt     = ""
    nbutwidth   = 0
    nbuttons    = 0
    nmax1       = 0
    nmax2       = 0
ELSE
    ndefault = 1
    DO CASE
        CASE UPPER(cprompt) == K_OK
            DIMENSION aprompt[1]
            nbuttons    = 1
            cprompt     = "\?\!\<OK"
            aprompt[1]  = "Ok"

        CASE UPPER(cprompt) == K_CONTINUE
            DIMENSION aprompt[1]
            nbuttons    = 1
            cprompt     = "\?\!\<Continue"
            aprompt[1]  = "Continue"

        CASE UPPER(cprompt) == K_OKCANCEL
            DIMENSION aprompt[2]
            IF "C" $ cprompt
                ndefault = 2
            ENDIF
            cprompt     = "\!\<OK;\?\<Cancel"
            nbuttons    = 2
            aprompt[1]  = "Ok"
            aprompt[2]  = "Cancel"

        CASE LEFT(UPPER(cprompt), 2) == K_YESNO
            DIMENSION aprompt[2]
            ctemp       = "\<Yes;\<No"
            aprompt[1]  = "Yes"
            aprompt[2]  = "No"

            IF UPPER(cprompt) == K_YESNOCANCEL
                *
                * Make cancel button react to ESC-key press
                *
                DIMENSION aprompt[3]
                nbuttons    = 3
                ctemp       = ctemp + ";\?\<Cancel"
                aprompt[3]  = "Cancel"
            ELSE
                nbuttons    = 2
            ENDIF

            DO CASE
                CASE "Y" $ cprompt
                    cprompt = "\!" + ctemp

                CASE "N" $ cprompt
                    cprompt  = STUFF(ctemp, 7, 0, "\!")
                    ndefault = 2

                CASE "C" $ cprompt
                    cprompt  = STUFF(ctemp, 12, 0, "\!")
                    ndefault = 3

                OTHERWISE
                    cprompt = ctemp
            ENDCASE

        CASE UPPER(cprompt) == K_RETRYCANCEL
            DIMENSION aprompt[2]
            IF "C" $ cprompt
                ndefault = 2
            ENDIF
            nbuttons    = 2
            cprompt     = "\!\<Retry;\?\<Cancel"
            aprompt[1]  = "Retry"
            aprompt[2]  = "Cancel"

        CASE UPPER(cprompt) == K_ABORTRETRYIGNORE
            DIMENSION aprompt[3]
            nbuttons    = 3
            ctemp       = "\<Abort;\<Retry;\<Ignore"
            aprompt[1]  = "Abort"
            aprompt[2]  = "Retry"
            aprompt[3]  = "Ignore"

            DO CASE
                CASE "A" $ cprompt
                    cprompt = "\!" + ctemp

                CASE "R" $ cprompt
                    cprompt  = STUFF(ctemp, 9, 0, "\!")
                    ndefault = 2

                CASE "I" $ cprompt
                    cprompt  = STUFF(ctemp, 17, 0, "\!")
                    ndefault = 3
            ENDCASE

        OTHERWISE
            *
            * Parse user-defined buttons
            *
            nbuttons  = OCCURS(";", cprompt) + 1
            nbutsize  = 0

            DIMENSION aprompt[nbuttons]

            lpicbut = get_prompt(cprompt, @aprompt, @nbuttons, @ndefault, ;
                                 @nbutheight, @nbutsize)

            *
            * If it's a picture button, then remove SIZE clause from the
            * prompt string.
            *
            IF lpicbut AND [SIZE] $ cprompt
                cprompt = ALLTRIM(LEFT(cprompt, AT("SIZE", cprompt) - 1))
            ENDIF
    ENDCASE

    nmaxbut = 0

    IF NOT lpicbut
        nbutsize = 0

        FOR i = 1 TO nbuttons
            IF nbutsize < TXTWIDTH(aprompt[i], cbfont, nbsize, cbstyle)
                nbutsize = TXTWIDTH(aprompt[i], cbfont, nbsize, cbstyle)
                nmaxbut  = i
            ENDIF
        ENDFOR
    ENDIF
ENDIF

*
* Determine if user wants window to be centered.
*
lcenter = (nrow = -1) OR (ncol = -1)
IF lcenter
    nrow = 0
    ncol = 0
ENDIF

*
* The width of the window includes:
*   1.  the icon width and gaps
*   2.  the longest message line or the length of the row of buttons
*   3.  the border between the text and right side of message window
*       (Indicated by "WW" in TXTWIDTH())
*
* If no icon is to be displayed, then the borders between text and the
* message window (Indicated by "WW" in TXTWIDTH()).
*
IF niconwidth <> 0
    nwwidth = nicongap * 2 + niconwidth + ;
        TXTWIDTH(amsg[nmaxstr]+"WW", cfont, nsize, cstyle)
ELSE
    *
    * Set icon gap to one character spacing if no icons are used.
    *
    nicongap= TXTWIDTH("W", cfont, nsize, cstyle)
    nwwidth = TXTWIDTH(amsg[nmaxstr]+"WWWW", cfont, nsize, cstyle)
ENDIF

*
* The height of the window includes...
*   1.  gap between 1st line and top of window
*   2.  gap between last line and buttons
*   3.  gap between buttons and bottom of window
*
IF NOT EMPTY(cprompt)
    ngap     = IIF(nmaxline < niconheight, 2.5, 3.5) * (nbfheight / nwfheight)
    nwheight = nmaxline + ngap + (nbutheight * nbfheight / nwfheight)

    IF (niconheight + (nbutheight * nbfheight / nwfheight) + ngap) > nwheight
        nwheight = niconheight + (nbutheight * nbfheight / nwfheight) + ngap
    ENDIF
ELSE
    nwheight = nmaxline + 2 + IIF(nmaxline > 1, 0, (nwfheight / nbfheight))

    IF (niconheight + 2) > nwheight
        nwheight = niconheight + 2
    ENDIF
ENDIF

*
* Redefine our window with the appropriate attributes
*
MODIFY WINDOW wmsgbox COLOR SCHEME 5 ;
    AT nrow,ncol SIZE nwheight,nwwidth ;
    TITLE ctitle FONT cfont,nsize STYLE cstyle ;
    SYSTEM FLOAT

*
* Now check this window to see if the buttons (if any) will fit.
*
IF NOT EMPTY(cprompt)
    *
    * Change the window's font attributes so we may position the button
    * list in the bottom-center of the window based on the button's font
    * properties.
    *
    MODIFY WINDOW wmsgbox FONT cbfont,nbsize STYLE cbstyle

    *
    * Determine the spaces between each button
    *
    DO CASE
        CASE nbuttons = 2
            ngap = 6
        CASE nbuttons = 3
            ngap = 4
        OTHERWISE
            ngap = 2
    ENDCASE

    *
    * Get the TXTWIDTH length of the longest button based on button font.
    *
    IF NOT lpicbut
        nbutsize  = TXTWIDTH(aprompt[nmaxbut], cbfont, nbsize, cbstyle) + 5
    ENDIF
    nbutwidth = (nbuttons * nbutsize) + (ngap * (nbuttons - 1))

    *
    * Now, make sure that the window is large enough to display all the
    * buttons.  If not, then resize to proper width and re-center the
    * buttons based on new window width.
    *
    nwwidth = (nbutwidth + ngap)
    IF (WCOLS() / nwwidth) <= 1
        *
        * Save current window height (based on new font info.)
        *
        nwheight = WROWS()

        MODIFY WINDOW wmsgbox AT nrow,ncol SIZE nwheight,nwwidth
    ENDIF

    *
    * Restore original window font so we can display our message
    *
    MODIFY WINDOW wmsgbox FONT cfont,nsize STYLE cstyle
ENDIF

IF lcenter
    MOVE WINDOW wmsgbox CENTER
ENDIF

*
* Fill the background with user specified color, if any.
*
IF TYPE("cncolor") == "N"
    @ 0,0 FILL TO WROWS(),WCOLS() COLOR SCHEME cncolor
ELSE
    @ 0,0 FILL TO WROWS(),WCOLS() COLOR (cncolor)
ENDIF

*
* Display the icon, centering vertically between the message area
*
IF NOT EMPTY(ciconfile)
    IF nmaxline < niconheight
        nsayrow = (1 + IIF(nmaxline > 1, 0, 0.5)) ;
                * (nbfheight / nwfheight) ;
                - (niconheight / (2 * (nmaxline + 1)))
        IF cbitmaps == [DBF]
            @ nsayrow,nicongap ;
                SAY EVAL(ciconfile) BITMAP SIZE niconheight,niconwidth
        ELSE
            @ nsayrow,nicongap SAY ciconfile BITMAP
        ENDIF
    ELSE
        nsayrow = (nmaxline - niconheight) / 2 ;
                + IIF(K_ICONHEIGHT > nwfheight, 0.75, 0.25)
        IF cbitmaps == [DBF]
            @ nsayrow, nicongap ;
                SAY EVAL(ciconfile) BITMAP SIZE niconheight,niconwidth
        ELSE
            @ nsayrow, nicongap SAY ciconfile BITMAP
        ENDIF
    ENDIF
ENDIF

*
* Caculate which row to start displaying the multi-line message
*
nsayrow = (1 + IIF(nmaxline > 1, 0, 0.5)) * (nbfheight / nwfheight)

DO CASE
    CASE cjust == K_LEFTJUST
        *
        * Set starting column for message lines at 1 character after icon
        *
        nsaycol = niconwidth + nicongap * 2

        IF TYPE("cncolor") == "N"
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR SCHEME cncolor
                nsayrow = nsayrow + 1
            ENDFOR
        ELSE
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR (amsgclr[i])
                nsayrow = nsayrow + 1
            ENDFOR
        ENDIF

    CASE cjust == K_CENTERJUST
        *
        * Set starting column for message lines at center of window
        * but after icon
        *
        IF niconwidth > 0
            nsaycol = niconwidth + nicongap ;
                    + (WCOLS() - niconwidth - 2 * nicongap)/2
        ELSE
            nsaycol = WCOLS()/2
        ENDIF

        IF TYPE("cncolor") == "N"
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol-TXTWIDTH(amsg[i],cfont,nsize,cstyle)/2;
                    SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR SCHEME cncolor
                nsayrow = nsayrow + 1
            ENDFOR
        ELSE
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol-TXTWIDTH(amsg[i],cfont,nsize,cstyle)/2;
                    SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR (amsgclr[i])
                nsayrow = nsayrow + 1
            ENDFOR
        ENDIF

    CASE cjust == K_RIGHTJUST
        *
        * Set starting column for message lines at right side of window
        * but 2 spaces before border (Indicated by "WW" in TXTWIDTH()).
        *
        nsaycol = WCOLS() - TXTWIDTH("WW", cfont, nsize, cstyle)

        IF TYPE("cncolor") == "N"
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol-TXTWIDTH(amsg[i],cfont,nsize,cstyle) ;
                    SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR SCHEME cncolor
                nsayrow = nsayrow + 1
            ENDFOR
        ELSE
            FOR i = 1 TO nmaxline
                @ nsayrow,nsaycol-TXTWIDTH(amsg[i],cfont,nsize,cstyle) ;
                    SAY amsg[i] ;
                    FONT cfont,nsize ;
                    STYLE cstyle ;
                    COLOR (amsgclr[i])
                nsayrow = nsayrow + 1
            ENDFOR
        ENDIF
ENDCASE

IF EMPTY(cprompt) OR (RDLEVEL() = K_MAXREADLEVEL)
    ACTIVATE WINDOW wmsgbox TOP
    IF (RDLEVEL() = K_MAXREADLEVEL)
        WAIT WINDOW "Maximum read level exceeded!!!" + CHR(13) + ;
            "Please contact " + K_CONTACTNAME NOWAIT
        = Beep()    
        = INKEY(0)
    ENDIF
ELSE
    *
    * Now change to window's font attribute to the button's font.
    *
    MODIFY WINDOW wmsgbox FONT cbfont,nbsize STYLE cbstyle

    *
    * Calculate the starting row and column for buttons...center of window
    *
    nsayrow = WROWS() - nbutheight - 1
    nsaycol = (WCOLS() - nbutwidth)/2
    cprompt = IIF(lpicbut, "*HB ", "*H ") + cprompt

    nchoice = ndefault
    DO WHILE .T.
        *
        * Create the buttons
        *
        @ nsayrow,nsaycol GET nchoice ;
            FUNCTION cprompt;
            SIZE nbutheight, nbutsize, ngap;
            FONT cbfont, nbsize ;
            STYLE cbstyle

        ACTIVATE WINDOW wmsgbox TOP
        = Beep()
        
        READ CYCLE ;
            MODAL OBJECT ndefault

        IF BETWEEN(nchoice, 1, nbuttons)
            retval = IIF(TYPE('nnumeric') == [N] AND nnumeric = 1, ;
                        nchoice, aprompt[nchoice])
            EXIT
        ENDIF
    ENDDO

    RELEASE WINDOW wmsgbox
ENDIF

*
* If no prompt is provided, then don't activate screen/window.
*
IF EMPTY(cwinsave)
    ACTIVATE SCREEN
ELSE
    IF lvisible
        ACTIVATE WINDOW (cwinsave) SAME
    ELSE
        ACTIVATE WINDOW (cwinsave) SAME NOSHOW
    ENDIF
ENDIF

*
* Restore environment settings
*
IF cdevice <> "SCREEN"
    SET DEVICE TO PRINTER
ENDIF

IF cconsole <> "ON"
    SET CONSOLE OFF
ENDIF

IF cprinter <> "OFF"
    SET PRINTER ON
ENDIF

FOR i = 1 TO 24
    IF (i <> 13) AND (i <> 14)
        SET COLOR OF SCHEME i TO (aclrset[i])
    ENDIF
ENDFOR

RETURN retval


***
*
*   GET_PROMPT()
*
*   Called by WMSGBOX() to parse the prompt string into an array of choices.
*
*   If more than one default settings was found, then the first one will be
*   the default one.
*
**
FUNCTION get_prompt
    PARAMETER cchoice, achoice, nmax, ndefault, nbutheight, nbutsize
    PRIVATE i, nloc, nloc2, lpicbut
    EXTERNAL ARRAY achoice

    ndefault = 0

    *
    * Picture buttons have .BMP or .ICO extensions for bitmapped objects, or
    * a SIZE clause!
    *
    lpicbut = ([.BMP] $ UPPER(cchoice)) OR ([.ICO] $ UPPER(cchoice)) OR ;
              ([SIZE] $ UPPER(cchoice))

    FOR i = 1 TO nmax
        nloc = AT(";", cchoice)

        IF nloc = 0
            IF lpicbut
                nloc        = AT("SIZE", UPPER(cchoice))
                achoice[i]  = IIF(nloc = 0, cchoice, LEFT(cchoice, nloc - 1))
            ELSE
                achoice[i]  = cchoice
            ENDIF
        ELSE
            achoice[i] = LEFT(cchoice, nloc - 1)
            cchoice    = SUBSTR(cchoice, nloc + 1)
        ENDIF

        IF ndefault = 0
            *
            * In order for a choice to be a default, it must have a "\!"
            * and it must not be right before a letter in the option string.
            *
            nloc2 = AT("\!", achoice[i])
            IF INLIST(nloc2, 1, 3)
                ndefault = i
            ENDIF
        ENDIF

        *
        * Now remove all control-character combinations from the string
        * and determine the longest button option.
        *
        achoice[i] = STRTRAN(STRTRAN(STRTRAN(achoice[i], '\!'), '\<'), '\?')
    ENDFOR

    IF lpicbut
        *
        * For, picture buttons, user must define the size of the buttons.
        *
        nloc    = AT("SIZE", UPPER(cchoice)) + 5
        nloc2   = AT(",", cchoice)

        *
        * If no size is given or invalid size, then assign default values...
        *
        IF nloc = 6 OR ((nloc2 - nloc) < 1)
            nbutheight  = 5
            nbutsize    = 10
        ELSE
            nbutheight  = VAL(ALLTRIM(SUBSTR(cchoice, nloc, (nloc2 - nloc))))
            nbutsize    = VAL(ALLTRIM(SUBSTR(cchoice, nloc2 + 1)))
        ENDIF
    ENDIF
RETURN lpicbut


***
*
*   GET_MSG()
*
*   Called by WMSGBOX() to parse the message string for embedded color
*   specifier.
*
*   If no color is specified, then default to RGBSCHEME(1,1), the color of
*   a user defined window.
*
*   'accolor' is passed by reference, therefore, what is assigned to it will
*   be returned to the calling procedure.
*
**
FUNCTION get_msg
    PARAMETER cmesg, accolor, nindex
    EXTERNAL ARRAY accolor
    PRIVATE nloc1, nloc2, ccolor

    nloc1 = AT('{', cmesg)
    nloc2 = AT('}', cmesg)

    IF nloc1 > 0 AND nloc2 > 0
        IF INLIST(SUBSTR(cmesg, nloc1, nloc2 - nloc1 + 1), ;
            K_BLACK, K_BLUE, K_CYAN, K_DRKBLUE, K_DRKCYAN, K_DRKGREEN, ;
            K_DRKMAGENTA, K_DRKRED, K_DRKYELLOW, K_DRKGRAY, K_GRAY, ;
            K_GREEN, K_MAGENTA, K_RED, K_WHITE, K_YELLOW)

            accolor[nindex] =  SUBSTR(cmesg, nloc1 + 1, nloc2-nloc1-1) ;
                            + SUBSTR(SET('color'), AT('/', SET('color')))
            cmesg           = LEFT(cmesg, nloc1 - 1)
        ELSE
            accolor[nindex] = cncolor
        ENDIF
    ELSE
        accolor[nindex] = cncolor
    ENDIF
RETURN cmesg

  


PROCEDURE AddMbar
   PARAMETER WindowName, Prompt
   
   PRIVATE i, x, foundit
   
   IF PARAMETERS() = 1
      m.prompt = PROPER(WTITLE(m.WindowName))
   ENDIF
   
   IF PARAMETERS() = 0
      m.WindowName = WONTOP()
      m.prompt = PROPER(WTITLE(m.WindowName))
   ENDIF
   
   m.prompt = PROPER(m.prompt)
   foundit = .f.
    
   FOR i = CNTBAR(m.gwindowpad) TO 1 STEP -1
 
      IF PRMBAR(m.gwindowpad,GETBAR(m.gwindowpad,i)) = m.prompt
         foundit = .t.
         EXIT
      ENDIF
      
   ENDFOR
      
   IF !foundit
      IF EMPTY(ga_Windows[1])
         x = 1
      ELSE   
         x = ALEN(ga_Windows) + 1
         DIMENSION ga_Windows[x]
      ENDIF
         
      ga_Windows[x] = m.prompt
      
      IF GETBAR(m.gwindowpad,CNTBAR(m.gwindowpad)) < 0
         i = CNTBAR(m.gwindowpad) + 1
      ELSE
         i = GETBAR(m.gwindowpad,CNTBAR(m.gwindowpad)) + 1
      ENDIF
   
      DEFINE BAR i OF (m.gwindowpad) ;
      PROMPT m.prompt ;
      AFTER _mlast
	  ON SELECTION BAR i OF (m.gwindowpad) ACTIVATE WINDOW &WindowName
   ENDIF

   RETURN
   
   
   
PROCEDURE RemMBar
   PARAMETER WindowName, Prompt
   
   PRIVATE i, x

   IF PARAMETERS() = 1
      m.prompt = PROPER(m.WindowName)
   ENDIF
   
   IF PARAMETERS() = 0
      m.WindowName = WONTOP()
      m.prompt = PROPER(WTITLE(m.WindowName))
   ENDIF
   
   m.prompt = PROPER(m.prompt)
   x = ASCAN(ga_Windows, m.prompt)
   
   IF x > 0
      = ADEL(ga_Windows, x)
      
      IF x > 1
         DIMENSION ga_Windows[x - 1]
      ELSE
         ga_Windows = ""
      ENDIF
         
   ENDIF
         
   FOR i = CNTBAR(m.gwindowpad) TO 1 STEP -1
 
      IF PRMBAR(m.gwindowpad, GETBAR(m.gwindowpad,i)) = m.prompt
         RELEASE BAR GETBAR(m.gwindowpad, i) OF (m.gwindowpad)
         EXIT
      ENDIF
      
   ENDFOR

   RELEASE WINDOW (WindowName)
   RETURN



FUNCTION Longdate
   * returns a long string of the current date
   
   PARAMETER date
   
   PRIVATE date, x
   
   SET CENTURY ON
   x = CDOW(m.date) + ' ' + MDY(m.date)
   SET CENTURY OFF
   RETURN (x)
   
   
   
PROCEDURE Standby
   * displays a message and forces a pause by a click or keypress
   * no value returned.  Message may be multi lines using the semicolon
   * to separate lines within the text.
   
   PARAMETERS Msg

    IF _WINDOWS
       Msg = STRTRAN(Msg, ";", CHR(13))
       Msg = STRTRAN(Msg, "#", CHR(13))
       STORE .F. TO m.font, m.f_size, m.f_style
       = Wmsgbox (msg, ' Activity Pause ', "!", "  \!\<Continue", -1, -1, "C", m.font, ;
           f_size, f_style, 7)
    ELSE
       Msg = STRTRAN(Msg, "#", ";")
       = Beep()
       DO Alert WITH Msg, "  \!\<Continue", "", " Standby Notice ", 7
    ENDIF
       
    RETURN
    
  
PROCEDURE P_setup
   * Standard setup procedure that can be used in ALL
   * setup snippets
   * include as first procedure in setup
   * (must be after the section header if there is one)
   *
   
   PRIVATE i, program
   
   IF _WINDOWS
      m.gnpushmenu = m.gnpushmenu + 1
   ELSE
      SET SYSMENU OFF
   ENDIF


   PUSH KEY CLEAR
   i = 1
   m.program = "x"
   
   DO WHILE !EMPTY(m.program)
      m.program = PROGRAM(i)
      
      IF m.program = "P_SETUP"
         i      = i - 1
         m.g_program = PROGRAM(i)
         EXIT
      ENDIF
         
      i         = i + 1
   ENDDO
   
   RETURN
   
 

PROCEDURE P_cleanup
   * Standard cleanup commands when exiting a screen
   WAIT CLEAR

   IF _WINDOWS
      m.gnpushmenu = m.gnpushmenu - 1
   ENDIF

   SET SYSMENU AUTOMATIC
   RETURN
   


PROCEDURE Alert
   * creates a window with string text (multiple lines in the str
   * should be separated by semicolons.
   * adds push buttons at the bottom with the value of the
   * selected push button returned in the action field
   

   * Strg     = Text of message that will be printed in alert box
   * cscheme  = Colour scheme to use, otherwise default is used
   * Buttons  = Buttons to be setup, string delimited by ;
   *            '\<Yes;\<No;\<Cancel' for example
   * Action   = the variable that will store the alert result
   *            this allows parameter passing to this routine with
   *            the result showing up in the variable in the calling program
   *
   * title    = title of the alert window
   *
   * cscheme  = colour scheme of window
   *
   * font     = (Windows only) font to use in window
   *
   * f_size   = (Windows only) font size (Numeric)
   *
   * f_style  = (Windows only) font style "B" "N" "I" "IB" etc
   *            default style is normal
   *
   * only the first three parameters are required
   *
   
   PARAMETERS strg, buttons, action, title, cscheme, font, f_size, f_style, bitmap
   
   PRIVATE dflts, i, vname, numlines, numbuts, remain, maxlen,;
           but_arry, top_row, bot_row, top_col, bot_col, in_cons, in_dev
   
   SET TALK OFF
   CLEAR TYPEAHEAD
   in_cons = SET('CONSOLE')
   in_dev  = SET('DEVICE')
   SET CONSOLE OFF
   SET DEVICE TO SCREEN
   
   numparams = 9
   DIMENSION dflts[numparams,2]
   dflts[1,1] = "STRG"
   dflts[2,1] = "BUTTONS"
   dflts[3,1] = "ACTION"
   dflts[4,1] = "TITLE"
   dflts[5,1] = "CSCHEME"
   dflts[6,1] = "FONT"
   dflts[7,1] = "F_SIZE"
   dflts[8,1] = "F_STYLE"
   dflts[9,1] = "BITMAP"
   dflts[1,2] = 'Press any key to continue'
   dflts[2,2] = 'Ok'
   dflts[3,2] = ''
   dflts[4,2] = ' System Alert '
   dflts[5,2] = 7
   dflts[6,2] = "Arial"
   dflts[7,2] = 10
   dflts[8,2] = "N"
   dflts[9,2] = "HAND"
   
   FOR i = 1 TO numparams

      IF PARAMETERS() < i
         vname = dflts[i,1]
         &vname = dflts[i,2]
      ENDIF

   ENDFOR

   IF EMPTY(m.title)
      m.title = dflts[4, 2]
   ENDIF
      
   m.numlines = OCCURS(';',strg) + 1          && number of message lines

   IF !EMPTY(buttons)
      m.numbuts  = OCCURS(';',buttons) + 1    && number of buttons
   ELSE
      m.numbuts = 0
   ENDIF

   DIMENSION alert_arry[numlines]

   IF numbuts > 0
      DIMENSION but_arry[numbuts]
   ELSE
      DIMENSION but_arry[1]
   ENDIF

   m.remain = m.strg
   m.maxlen = 0

   IF _WINDOWS
      Msg = STRTRAN(strg, "#", CHR(13))
      Msg = STRTRAN(strg, ";", CHR(13))
      m.action = Wmsgbox (msg, title, m.bitmap, buttons, -1, -1, "C", m.font, ;
        f_size, f_style)
      RETURN
      
   ELSE
      FOR i = 1 TO m.numlines

         IF AT(';',m.remain) > 0
            alert_arry[i] = SUBSTR(m.remain,1,AT(';',m.remain)-1)
            alert_arry[i] = CHRTRAN(alert_arry[i],';','')
            m.remain = SUBSTR(m.remain,AT(';',m.remain)+1)
         ELSE
            alert_arry[i] = m.remain
            m.remain = ''
         ENDIF
      
         IF LEN(alert_arry[i]) > SCOLS() - 6
            alert_arry[i] = SUBSTR(alert_arry[i],1,SCOLS()-6)
         ENDIF

         IF LEN(alert_arry[i]) > m.maxlen
            m.maxlen = LEN(alert_arry[i])
         ENDIF

      ENDFOR

      m.remain = m.buttons
      m.maxbut = 0

      FOR i = 1 TO m.numbuts

          IF AT(';',m.remain) > 0
             but_arry[i] = SUBSTR(m.remain,1,AT(';',m.remain)-1)
             but_arry[i] = CHRTRAN(but_arry[i],';','')
             m.remain = SUBSTR(m.remain,AT(';',m.remain)+1)
         ELSE
             but_arry[i] = m.remain
             m.remain = ''
         ENDIF

         IF LEN(but_arry[i]) > m.maxbut
            m.maxbut = LEN(but_arry[i])
         ENDIF

      ENDFOR

      IF !EMPTY(buttons)
         m.numlines = m.numlines + 1
      ENDIF

      m.butsize = m.numbuts * (m.maxbut + 5) - 3   && chevrons + space

      IF m.butsize > m.maxlen
         m.maxlen = m.butsize
      ENDIF

      m.top_row = INT( (SROWS() - 4 - m.numlines) / 2)
      m.bot_row = m.top_row + 3 + m.numlines

      m.top_col = INT((SCOLS() - m.maxlen - 6) / 2)
      m.bot_col = m.top_col + m.maxlen + 6

      = W_grow("alert", m.top_row, m.top_col, m.bot_row,m.bot_col,;
               "F", m.title, cscheme)
      ACTIVATE WINDOW alert
      DO HighLight

      IF !EMPTY(buttons) AND RDLEVEL() < 5

         FOR i = 1 TO m.numlines - 1
            @ i,3 SAY PADC(alert_arry[i],m.maxlen)
         ENDFOR

         m.butpos = 3 + (m.maxlen - m.butsize)/2
         @ m.numlines + 1, m.butpos GET m.action FUNCTION '*H' PICTURE buttons;
                 SIZE 1,m.maxbut+2, 3
         READ MODAL

      ELSE

         FOR i = 1 TO m.numlines
            @ i,3 SAY PADC(alert_arry[i],m.maxlen)
         ENDFOR

         SET CONSOLE OFF
         =INKEY(0,'HM')
         SET CONSOLE ON


      ENDIF

   ENDIF

   RELEASE WINDOW alert

   IF m.in_cons = "OFF"
      SET CONSOLE OFF
   ENDIF

   IF m.in_dev = "PRINT"
      SET DEVICE TO PRINT
   ENDIF

   RETURN


PROCEDURE Close_msg
   * used to close any window opened with msg2user
   
   DEACTIVATE WINDOW Msg2user
   RELEASE WINDOW Msg2user
   RETURN



FUNCTION ready2pr
  * returns true or false depending on whether the printer is working
  
  DO WHILE .NOT. PRINTSTATUS()

    IF .NOT. YESNO(' Printer not ready! Try Again...')
      RETURN .f.
    ENDIF

  ENDDO

  RETURN .t.
  

FUNCTION Yesno
   * returns true or false depending on whether the user pressed
   * yes or no to the prompt.
   *
   * msg	the message or question that requires a yes or no response.
   *
   
   PARAMETERS Msg
   PRIVATE Button
   
   Msg = STRTRAN(Msg, "#", ";")
   m.action = "Yes"
   DO Alert WITH Msg,"\!\<Yes;\?\<No", m.action, "", 5

   RETURN IIF(m.action = "Yes", .t., .f.)
   
   

FUNCTION Esc
    RETURN IIF(LASTKEY() == 27, .T., .F.)


FUNCTION Get_rec
   * returns 0 or the current record number
   RETURN IIF(EOF() OR RECNO() > RECCOUNT(), 0, RECNO())



FUNCTION Rest_rec
   * restores the record pointer to the record number passed
   * If the mode is 1, it will also unlock the record
   * mode parameter to allow for slight differences in requirements
   * when the previous record is returned to.
   
   PARAMETER recno, mode

   PRIVATE recno, mode
   
   IF TYPE ("m.mode") = "N"
      DO CASE
         CASE m.mode = 1
              * basic mode usually done on a cancel
              
              IF BETWEEN(m.recno, 1, RECCOUNT())
                 GOTO m.recno
                 UNLOCK
                 SCATTER MEMVAR
                 RETURN .T.
              ELSE
                 RETURN .F.
              ENDIF   
      
      ENDCASE
          
   ELSE
   
      IF BETWEEN(m.recno, 1, RECCOUNT())
         GOTO recno
      ENDIF
   
   ENDIF
   
   RETURN recno


*
* STRIPPATH - Strip the path from a file name.
*
* Description:
* Find positions of backslash in the name of the file.  If there is one
* take everything to the right of its position and make it the new file
* name.  If there is no slash look for colon.  Again if found, take
* everything to the right of it as the new name.  If neither slash
* nor colon are found then return the name unchanged.
*
* Parameters:
* filename - character string representing a file name
*
* Return value:
* The string "filename" with any path removed
*
FUNCTION Strippath

   PARAMETER m.filename

   PRIVATE m.slashpos, m.namelen, m.colonpos

   m.slashpos = RAT("\", m.filename)

   IF m.slashpos <> 0
	  m.namelen  = LEN(m.filename) - m.slashpos
      m.filename = RIGHT(m.filename, m.namelen)
   ELSE
	  m.colonpos = RAT(":", m.filename)

      IF m.colonpos <> 0
		 m.namelen  = LEN(m.filename) - m.colonpos
		 m.filename = RIGHT(m.filename, m.namelen)
	  ENDIF
	  
   ENDIF
   
   RETURN m.filename


PROCEDURE Beep
   * a procedure to sound the speaker.
   
    SET BELL TO 2000,5
    ?? CHR(7)
    SET BELL TO
    RETURN
    


FUNCTION Fxuse
   * Selects the file if in use
   * otherwise uses it

   * m.alias 	name of the DBF file to use
   *
   * m.tag      name of the tag to set the order to
   *
   * m.key		key of the tag. If the tag does not exist, this key
   *			will be used to create it. (Not used much, or at all)
   
   
   PARAMETERS m.alias, m.tag, m.key
   
   PRIVATE m.ext, m.fullname
   m.ext      = .F.
   m.fullname = ''
   
   IF !(ALIAS() == m.alias)
   
      IF USED(m.alias)
         SELECT (m.alias)
      ELSE
         SELECT 0
         
         * Mimic FoxPro, allow extension or not
         IF '.' $ RIGHT(m.alias, 4)
            m.fullname = m.alias
            m.ext      = .T.
         ELSE
            m.fullname = m.alias + '.dbf'
         ENDIF
         
         IF FILE(fullname)
            USE(m.fullname)
         ELSE
            WAIT "File '" + m.fullname + "' Not Found" WINDOW
            RETURN ''
         ENDIF
      ENDIF
   ENDIF
   
   * was a Tag Passed?
   IF !EMPTY(m.tag)
      = Fxcdx(M.alias, m.tag, m.key)
   ENDIF
   
   RETURN ALIAS()
   

   
FUNCTION Fxcdx
   *
   * Check for the existence of the specified TAG,
   * or create it if necessary
   
   PARAMETERS m.alias, m.tag, m.key
   
   m.alias = UPPER(m.alias)
   m.tag   = UPPER(m.tag)
   m.key   = IIF(EMPTY(m.key),'', UPPER(m.key))
   
   PRIVATE i
   
   *
   * File must be in use in current area
   *
   
   i = 1
   
   IF !EMPTY(TAG(m.alias, i))
   
      DO WHILE !EMPTY(TAG(m.alias, i))
         
         IF TAG(m.alias, i) == m.tag
            m.key = KEY(i)
            i = 0
            EXIT
         ELSE
            i = i + 1
         ENDIF
         
      ENDDO
   ENDIF
   
   IF i !=0 AND !EMPTY(m.tag)
      
      IF EMPTY(m.tag)
         WAIT 'Tag not found. Key expression not passed' WINDOW
         RETURN
      ELSE
         IF !EMPTY(m.key)
            SET TALK ON
            SET TALK WINDOW
            INDEX ON &key TAG (m.tag)
            SET TALK OFF
         ELSE
            WAIT WINDOW 'No key expression passed. Order not set.'
            RETURN   
         ENDIF   
      ENDIF
   ENDIF
   
   SET ORDER TO (m.tag)
   
   RETURN      
                           
   
FUNCTION w_grow
   * this function will "grow" a box to size
   * 
   * w_name		name of the window to grow
   *
   * upper_row  upper row coordinate
   *
   * left_col	left column coordinate
   *
   * low_row 	lower row coordiante
   *
   * rt_col		right column coordinate
   *
   * box_kind	panel kind, D for double, P for panel Etc
   *
   * title		title of the window
   *
   * c_scheme	color scheme to use
   *  w_grow(<expC1>,<expN1>,<expN2>,<expN3>,<expN4>;       *
   *           [,<expC2>][,<expC3>[,<expN5>]])              *
   
   PARAMETERS w_name, upper_row, left_col, low_row, rt_col, box_kind;
           title, c_scheme

   PRIVATE w_name, upper_row, left_col, low_row, rt_col, box_kind, ;
           box_kind, c_scheme, mid_row, mid_col, r_step, c_step, ;
           sizer, i, params, x, y, one, two, stepper

   *  w_grow(<expC1>,<expN1>,<expN2>,<expN3>,<expN4>;       *
   *           [,<expC2>][,<expC3>[,<expN5>]])              *

   params = parameters()         &&  how many were passed
   one = 1
   two = 2


   IF params = 5
      box_kind = ''
      c_scheme = 8
   ENDIF

   IF params = 6

      IF type('box_kind') = 'C'

         IF len(box_kind) < 3
            box_kind = upper(box_kind)
            m.title = ''
            c_scheme = 8
         ELSE
            m.title = box_kind
            box_kind = ''
            c_scheme = 8
         ENDIF

     ELSE

         IF type('box_kind') = 'N'
            box_kind = ''
            m.title = ''
            c_scheme = box_kind
         ENDIF

      ENDIF

   ENDIF

   IF params = 7

     IF type('m.title') = 'N'
       c_scheme = m.title
       m.title = ''
     ENDIF

   ENDIF

   box_kind = upper(box_kind)

   DO CASE

      CASE box_kind = 'D'
           box = 'DOUBLE'

      CASE box_kind = 'P'
           box = 'PANEL'

       CASE box_kind = 'F'
           box = 'FLOAT'
           m.box_kind = " "
      OTHERWISE
           box = ''

   ENDCASE

   IF RIGHT(box_kind, one) = "F"
      m.box = m.box + " FLOAT"
   ENDIF
      
    IF _WINDOWS
         IF !EMPTY(m.title)
            m.title = 'TITLE "' + m.title + '"'
         ENDIF

       DEFINE WINDOW (w_name) FROM upper_row, left_col TO low_row, rt_col ;
                     &box &title COLOR SCHEME c_scheme

      ACTIVATE WINDOW (w_name)
      RETURN
   ENDIF
     
   ***  first find the screen coordinates for the middle of the
   ***  box for both row and column

   mid_row = upper_row + int((low_row - upper_row)/two)
   mid_col = left_col  + int((rt_col - left_col)/two)

   r_step = int((low_row - upper_row))
   c_step = int((rt_col - left_col))
   sizer  = iif(r_step > c_step, r_step, c_step)
   r_step = r_step/(sizer)
   c_step = c_step/(sizer)

   DEFINE WINDOW tmp FROM upper_row, left_col TO low_row, rt_col ;
                 NONE ZOOM COLOR SCHEME c_scheme
   ***  explode the window in steps
   x = 0
   y = 0
   stepper = CEILING(AT(SYS(17), "80486 80386 80286")/3) + 1
   
   FOR i = one TO sizer STEP m.stepper
       x = x + (r_step * m.stepper)
       y = y + (c_step * m.stepper)

       ZOOM WINDOW tmp NORM;
            AT mid_row - x/two + one, mid_col - y/two + one;
            SIZE x, y
       SHOW WINDOW tmp

   ENDFOR

   ***  now activate the real window!

   IF !EMPTY(m.title)
      m.title = 'TITLE "' + m.title + '"'
   ENDIF

   DEFINE WINDOW (w_name) FROM upper_row, left_col TO low_row, rt_col ;
                  &box &title COLOR SCHEME c_scheme

   * Remove the temporary window
   ACTIVATE WINDOW (w_name)
   RELEASE WINDOW tmp
   RETURN



FUNCTION Highlight
   * When the last entry was by a mouse click, the current object
   * is not high lighted. This function will high light the field
   * so the user can see where the cursor current is.

   IF _DOS
      CLEAR TYPEAHEAD
      KEYBOARD " "     && Force last event to Keyboard
      = INKEY(0)
   ENDIF

   RETURN



*
* ACTTHERM(<text>) - Activate thermometer.
*
* Activates thermometer.  Update the thermometer with UPDTHERM().
* Thermometer window is named "thermometer."  Be sure to RELEASE
* this window when done with thermometer.  Creates the global
* m.g_thermwidth.
*
*!*****************************************************************************
*!
*!      Procedure: ACTTHERM
*!
*!*****************************************************************************
PROCEDURE acttherm
   PARAMETER m.prompt
   
   PRIVATE m.text
   PUBLIC m.recthresh, g_thermwidth, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle

   m.g_dlgface	 = IIF(_MAC,"Geneva","MS Sans Serif")
   m.g_dlgsize	 = IIF(_MAC,10,8.000)
   m.g_dlgstyle  = IIF(_MAC,"","B")

   m.recthresh    = 1        && show it if more than this many input records
   m.g_thermwidth = 0		 && Thermometer width

   m.text = ""
   
   IF _MAC OR _WINDOWS
 
      IF TXTWIDTH(m.prompt, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle) > 43
    
         DO WHILE TXTWIDTH(m.prompt+"...", m.g_dlgface, m.g_dlgsize, m.g_dlgstyle) > 43
            m.prompt = LEFT(m.prompt, LEN(m.prompt)-1)
         ENDDO
      
         m.prompt = m.prompt + "..."
      ENDIF
   
      DO CASE

         CASE _WINDOWS
              DEFINE WINDOW thermomete ;
                AT  INT((SROW() - (( 5.615 * ;
                FONTMETRIC(1, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle )) / ;
                FONTMETRIC(1, WFONT(1,""), WFONT( 2,""), WFONT(3,"")))) / 2), ;
                INT((SCOL() - (( 63.833 * ;
                FONTMETRIC(6, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle )) / ;
                FONTMETRIC(6, WFONT(1,""), WFONT( 2,""), WFONT(3,"")))) / 2) ;
                SIZE 5.615,63.833 ;
                FONT m.g_dlgface, m.g_dlgsize ;
                STYLE m.g_dlgstyle ; 
                NOFLOAT ;
                NOCLOSE ;
                NONE ;
                COLOR RGB(0, 0, 0, 192, 192, 192)
              MOVE WINDOW thermomete CENTER
              ACTIVATE WINDOW thermomete NOSHOW

              @ 0.5,3 SAY m.text FONT m.g_dlgface, m.g_dlgsize STYLE m.g_dlgstyle
              @ 1.5,3 SAY m.prompt FONT m.g_dlgface, m.g_dlgsize STYLE m.g_dlgstyle
              @ 0.000,0.000 TO 0.000,63.833 ;
                 COLOR RGB(255, 255, 255, 255, 255, 255)
              @ 0.000,0.000 TO 5.615,0.000 ;
                 COLOR RGB(255, 255, 255, 255, 255, 255)
              @ 0.385,0.667 TO 5.231,0.667 ;
                 COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 0.308,0.667 TO 0.308,63.167 ;
                COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 0.385,63.000 TO 5.308,63.000 ;
                COLOR RGB(255, 255, 255, 255, 255, 255)
              @ 5.231,0.667 TO 5.231,63.167 ;
                COLOR RGB(255, 255, 255, 255, 255, 255)
              @ 5.538,0.000 TO 5.538,63.833 ;
                COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 0.000,63.667 TO 5.615,63.667 ;
                COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 3.000,3.333 TO 4.231,3.333 ;
                COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 3.000,60.333 TO 4.308,60.333 ;
                COLOR RGB(255, 255, 255, 255, 255, 255)
              @ 3.000,3.333 TO 3.000,60.333 ;
                COLOR RGB(128, 128, 128, 128, 128, 128)
              @ 4.231,3.333 TO 4.231,60.500 ;
                COLOR RGB(255, 255, 255, 255, 255, 255)
              m.g_thermwidth = 56.269
              
         CASE _MAC
              DEFINE WINDOW thermomete ;
                AT  INT((SROW() - (( 5.62 * ;
                FONTMETRIC(1, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle )) / ;
                FONTMETRIC(1, WFONT(1,""), WFONT( 2,""), WFONT(3,"")))) / 2), ;
                INT((SCOL() - (( 63.83 * ;
                FONTMETRIC(6, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle )) / ;
                FONTMETRIC(6, WFONT(1,""), WFONT( 2,""), WFONT(3,"")))) / 2) ;
                SIZE 5.62,63.83 ;
                FONT m.g_dlgface, m.g_dlgsize ;
                STYLE m.g_dlgstyle ;
                NOFLOAT ;
                NOCLOSE ;
		    	NONE ;
                COLOR RGB(0, 0, 0, 192, 192, 192)
              MOVE WINDOW thermomete CENTER
              ACTIVATE WINDOW thermomete NOSHOW

              IF ISCOLOR()
                 @ 0.000,0.000 TO 5.62,63.83 PATTERN 1;
                 COLOR RGB(192, 192, 192, 192, 192, 192)
                 @ 0.000,0.000 TO 0.000,63.83 ;
                 COLOR RGB(255, 255, 255, 255, 255, 255)
                 @ 0.000,0.000 TO 5.62,0.000 ;
         	     COLOR RGB(255, 255, 255, 255, 255, 255)
      	         @ 0.385,0.67 TO 5.23,0.67 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 0.31,0.67 TO 0.31,63.17 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 0.385,63.000 TO 5.31,63.000 ;
         	     COLOR RGB(255, 255, 255, 255, 255, 255)
      	         @ 5.23,0.67 TO 5.23,63.17 ;
         	     COLOR RGB(255, 255, 255, 255, 255, 255)
      	         @ 5.54,0.000 TO 5.54,63.83 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 0.000,63.67 TO 5.62,63.67 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 3.000,3.33 TO 4.23,3.33 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 3.000,60.33 TO 4.31,60.33 ;
         	     COLOR RGB(255, 255, 255, 255, 255, 255)
      	         @ 3.000,3.33 TO 3.000,60.33 ;
         	     COLOR RGB(128, 128, 128, 128, 128, 128)
      	         @ 4.23,3.33 TO 4.23,60.33 ;
         	     COLOR RGB(255, 255, 255, 255, 255, 255)
              ELSE
                 @ 0.000, 0.000 TO 5.62, 63.830  PEN 2
                 @ 0.230, 0.500 TO 5.39, 63.333  PEN 1
	          ENDIF
	          
              @ 0.5,3 SAY m.text FONT m.g_dlgface, m.g_dlgsize STYLE m.g_dlgstyle+"T" ;
              COLOR RGB(0,0,0,192,192,192)
              @ 1.5,3 SAY m.prompt FONT m.g_dlgface, m.g_dlgsize STYLE m.g_dlgstyle+"T" ;
              COLOR RGB(0,0,0,192,192,192)

              m.g_thermwidth = 56.27
		
		      IF !ISCOLOR()
   		         @ 3.000,3.33 TO 4.23,m.g_thermwidth + 3.33 
		      ENDIF
      ENDCASE
   
      SHOW WINDOW thermomete TOP
   ELSE

      DEFINE WINDOW thermomete;
         FROM INT((SROW()-6)/2), INT((SCOL()-57)/2) ;
         TO INT((SROW()-6)/2) + 6, INT((SCOL()-57)/2) + 57;
         DOUBLE COLOR SCHEME 5
      ACTIVATE WINDOW thermomete NOSHOW

      m.g_thermwidth = 50
      @ 0,3 SAY m.prompt
      @ 2,1 TO 4,m.g_thermwidth+4

      SHOW WINDOW thermomete TOP
   ENDIF
   
   RETURN

*
* UPDTHERM(<percent>) - Update thermometer.
*
*!*****************************************************************************
*!
*!      Procedure: UPDTHERM
*!
*!      Called by: BUILD              (procedure in GENSCRN.PRG)
*!               : DISPATCHBUILD      (procedure in GENSCRN.PRG)
*!               : BUILDCTRL          (procedure in GENSCRN.PRG)
*!               : EXTRACTPROCS       (procedure in GENSCRN.PRG)
*!               : BUILDFMT           (procedure in GENSCRN.PRG)
*!
*!*****************************************************************************
PROCEDURE updtherm
   PARAMETER m.percent
   
   PRIVATE m.nblocks, m.percent

   IF !WEXIST("thermomete")
      DO acttherm WITH "Updating ..."
   ENDIF

   ACTIVATE WINDOW thermomete

   m.nblocks = (m.percent/100) * (m.g_thermwidth)

   DO CASE
      CASE _DOS
           @ 3,3 SAY REPLICATE("",m.nblocks)
      
      CASE _WINDOWS
           @ 3.000,3.333 TO 4.231,m.nblocks + 3.333 ;
               PATTERN 1 COLOR RGB(128, 128, 128, 128, 128, 128)
               
      CASE _MAC
           @ 3.000,3.33 TO 4.23,m.nblocks + 3.33 ;
           PATTERN 1 COLOR RGB(0, 0, 128, 0, 0, 128)
           
      OTHERWISE
           @ 3,3 SAY REPLICATE("",m.nblocks)
           
   ENDCASE
   
   RETURN

*
* DEACTTHERMO - Deactivate and Release thermometer window.
*
*!*****************************************************************************
*!
*!      Procedure: DEACTTHERMO
*!
*!*****************************************************************************
PROCEDURE deactthermo

   RELEASE m.recthresh, g_thermwidth, m.g_dlgface, m.g_dlgsize, m.g_dlgstyle

   IF WEXIST("thermomete")
      RELEASE WINDOW thermomete
   ENDIF
   
RETURN


PROCEDURE Movetherm

   DO CASE
      CASE RECCOUNT() > 1000

         IF RECNO() % 100 = 0
            DO updtherm WITH INT(RECNO() / RECCOUNT() * 100)
         ENDIF

      OTHERWISE

         IF RECNO() % 10  = 0
            DO updtherm WITH INT(RECNO() / RECCOUNT() * 100)
         ENDIF

      ENDCASE

